import inspect
import json
import os
from collections import OrderedDict
from functools import wraps
from pathlib import Path
from typing import Callable

from .utils import current_time, sort_times

DEFAULT_ROLE = "user"
MESSAGE_TYPE = dict[str, str]
MESSAGES_TYPE = list[MESSAGE_TYPE] | None
# MESSAGES_TYPE = list[dict[str, str]]


def chat_message(role: str, content: str) -> dict[str, str]:
    return {"role": role, "content": content}


def system_message(content: str) -> dict[str, str]:
    return chat_message(role="system", content=content)


def user_message(content: str) -> dict[str, str]:
    return chat_message(role="user", content=content)


def assistant_message(content: str) -> dict[str, str]:
    return chat_message(role="assistant", content=content)


def str_to_messages(messages: MESSAGES_TYPE) -> MESSAGES_TYPE:
    messages = [messages] if not isinstance(messages, list) else messages
    for i, message in enumerate(messages):
        if isinstance(message, str):
            messages[i] = user_message(content=message)
        elif not isinstance(message, dict):
            raise TypeError(f"Invalid message type: {type(message)}")
    return messages


def process_messages(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        kwargs = inspect.getcallargs(func, *args, **kwargs)
        if not (messages := kwargs.get("messages")):
            return []
        messages = str_to_messages(messages)
        kwargs["messages"] = messages
        return func(**kwargs)

    original_signature = inspect.signature(func)
    wrapper.__annotations__ = {
        k: v.annotation
        for k, v in original_signature.parameters.items()
        if v.annotation is not inspect.Parameter.empty
    }
    if original_signature.return_annotation is not inspect.Signature.empty:
        wrapper.__annotations__["return"] = original_signature.return_annotation

    return wrapper


def last_message(
    messages: MESSAGES_TYPE, role: str = ""
) -> dict[str, int | dict[str, str]]:
    """Returns the last message with the given role."""

    if role:
        messages = [
            [i, message]
            for i, message in enumerate(messages)
            if message["role"] == role
        ]
    if len(messages) == 0:
        return None
    else:
        return {"index": messages[-1][0], "message": messages[-1][1]}


def last_user_message(messages: MESSAGES_TYPE) -> dict:
    message = last_message(messages, role="user")
    # print(f"Last user message: {message}")
    return message


def last_assistant_message(messages: MESSAGES_TYPE) -> dict:
    message = last_message(messages, role="assistant")
    # print(f"Last assistant message: {message}")
    return message


# @process_messages
def merge_same_role_messages(messages: MESSAGES_TYPE) -> MESSAGES_TYPE:
    if not messages:
        return []
    new_messages = []
    last_message = None
    for message in messages:
        if last_message is None:
            last_message = message
        elif last_message["role"] == message["role"]:
            last_message["content"] += "\n\n" + message["content"]
        else:
            new_messages.append(last_message)
            last_message = message
    if last_message is not None:
        new_messages.append(last_message)
    return new_messages


# @process_messages
def del_messages(messages: MESSAGES_TYPE, idx: int | list[int] = -1) -> MESSAGES_TYPE:
    """Delete the messages at the given indices."""
    if not messages:
        return []
    if not isinstance(idx, list):
        idx = [idx]
    idx = [i % len(messages) for i in idx if i < len(messages)]
    return [message for i, message in enumerate(messages) if i not in idx]


# @process_messages
def merge_n_messages(
    messages: MESSAGES_TYPE = None,
    merged_role: str = DEFAULT_ROLE,
    n_messages: int = 2,
) -> MESSAGES_TYPE:
    """Merge the last n_messages into a single message with the given role."""

    if len(messages) < n_messages:
        return messages
    last_n_messages = messages[-n_messages:]
    merged_content = "\n\n".join([m["content"] for m in last_n_messages])
    messages[-n_messages:] = [{"role": merged_role, "content": merged_content}]
    return messages


# @process_messages
def edit_messages(
    messages: MESSAGES_TYPE,
    idx: int | list[int] = -1,
    content: str | list[str] | None = None,
    role: str | list[str] | None = None,
) -> MESSAGES_TYPE:
    """Edit the messages at the given indices."""
    if not messages:
        return []
    if not isinstance(idx, list):
        idx = [idx]
    idx = [i % len(messages) for i in idx if i < len(messages)]
    for i, id in enumerate(idx):
        if content:
            if isinstance(content, list):
                messages[id]["content"] = content[i]
            else:
                messages[id]["content"] = content
        if role:
            if isinstance(role, list):
                messages[id]["role"] = role[i]
            else:
                messages[id]["role"] = role
    return messages


# @process_messages
def set_role(messages: MESSAGES_TYPE, role: str = DEFAULT_ROLE) -> MESSAGES_TYPE:
    """Set the role of the last message."""
    if not messages:
        return []
    messages[-1]["role"] = role
    return messages


def set_user_role(messages: MESSAGES_TYPE) -> MESSAGES_TYPE:
    return set_role(messages=messages, role="user")


def set_assistant_role(messages: MESSAGES_TYPE) -> MESSAGES_TYPE:
    return set_role(messages=messages, role="assistant")


def messages_to_chat(
    messages: MESSAGES_TYPE, chat_name: str = ""
) -> dict[str, list[dict[str, str]]]:
    chat_name = chat_name or current_time()
    return {chat_name: messages}


def save_chats(chats: dict[str, list[dict[str, str]]], chats_dir: str | Path = "chats"):
    os.makedirs(chats_dir, exist_ok=True)
    for chat_name, messages in chats.items():
        with open(f"{chats_dir}/{chat_name}.json", "w") as f:
            json.dump(messages, f, indent=4)


def load_chats(chats_dir: str | Path = "chats") -> dict[str, list[dict[str, str]]]:
    os.makedirs(chats_dir, exist_ok=True)
    chats = {}
    for file in Path(chats_dir).glob("*.json"):
        chat_name = Path(file).stem
        messages = json.load(file)
        chats[chat_name] = messages
    try:
        chats = OrderedDict(
            {chat_name: chats[chat_name] for chat_name in sort_times(chats.keys())}
        )
    except Exception:
        pass
    return chats



from ..utils import deindent

pre_improve_prompt = deindent(
    """
I get help from a bunch of AI assistants to do different tasks like generating lectures, asking random questions, solving math problems, retreiving knowledge and more.
I don't always come up with the best prompts. So I have to do some back and forth with the assistants to get the best results.
You know the best prompt engineering techniques and as an AI assistant yourself, you can help me with this.
Suppose I have written a prompt and I want it to be improved, how should I ask an AI assistant to improve it? Assume that the assistant also knows the prompt.
Can I get a generic answer that I can just copy paste regardless of the original prompt?
"""
)

improve_prompt = deindent(
    """
I need your help.
I need you to act as a Chatbot Architect with a deep understanding of prompt engineering for GPTs by Open AI.

The latest message above is a prompt for an AI assistant.
While the prompt serves its basic function, I want to ensure that it's optimized for the best possible outcome. Could you please help me refine it? Specifically, I'm looking to:

1. Make the prompt more clear and specific, so the AI understands exactly what I'm asking for.
2. Ensure that it's directed towards generating the most relevant and high-quality content/output.
3. Embed any necessary constraints or guidelines that would hone the responses to be more in line with my objectives.

Your prompt engineering expertise would be greatly appreciated.
I will use your prompt as it is, so please don't add any comments or notes or prefacing or ending statements.
Thank you!
"""
)

gen_rag_query_prompt = deindent(
    """
You are an expert at world knowledge.
I need a query for Retreival Augmented Generation (RAG) based on the messages above. The messages are part of an ongoing conversation so they may have some unnecessary text like "Okay so tell me about...", "I want to know about...", "I'm curious about...", etc.
Give me an effective query for RAG by optimizing it for similarity search in a vector database like ChromaDB, FAISS, etc.
I will use your query as it is, so please don't add any comments or notes or prefacing or ending statements.
Thank you!
"""
)

gen_web_search_query_prompt = deindent(
    """
You are an expert at web search.
I need a web search query based on the messages above. The messages are part of an ongoing conversation so they may have some unnecessary text like "Okay so tell me about...", "I want to know about...", "I'm curious about...", etc.
Give me an effective query for a web search engine like Google, Bing, DuckDuckGo, etc.
I will use your query as it is, so please don't add any comments or notes or prefacing or ending statements.
Thank you!
"""
)


path_selection_prompt = deindent(
    """
*Objective*: You are an AI assistant required to understand the context and nuance of a human conversation and select the most appropriate predefined sequence of tasks (referred to as "paths") that aligns with the conversation's implied intent. You will receive conversation history and a series of path descriptions. Your job is to parse the information, deduce the correct context, and identify which path best corresponds to the conversation's implicit or explicit request.

*Instructions*: Follow these steps every time you analyze a conversation history:

1. Carefully read the conversation history provided to understand the tone, mood, and content of the dialogue.
2. Assess previous messages and look for contextual clues that might influence the path selection.
3. Examine the path descriptions to understand the tasks that each path will execute.
4. Determine which path best complements the conversation history, based on relevance and appropriateness.
5. Reply with a JSON object containing the index of the path you determine to be the best fit.

*Example*:

## Conversation History ##
- "I'm eager to kick back and relax."
- "Yeah, it's been a long day. Let's make it cozy in here."

## Path Descriptions ##
0. "Close curtains and turn off lights" - This path involves creating a darker and more intimate atmosphere which could be suitable for resting or creating a cozy environment.
1. "Turn on lights and open curtains" - This path involves letting in more light and is typically more suitable for activities that require bright lighting or enjoying a view.

*Decision Criteria*:
Given the hints in the conversation about wanting to relax and make the environment cozy, you can infer that a darker and more relaxing environment is desired. Thus, the appropriate path would be the one that involves closing the curtains and turning off the lights.

*Correct Response*:
```json
{"path_index": 0}
```

Now it's your turn. Analyze the conversation and select the most appropriate path.
You'll get a cookie if you get it right! And I know you like cookies.
"""
)

lc_code_prompt = deindent(
    """
You are an expert programmer and problem-solver, tasked with answering any question
about Langchain and generating complete working scripts using the Langchain framework. Your responses
shall have two parts: Answer and Code.

For the Answer: Generate a comprehensive and informative answer of 80 words or less for the
given question based solely on the provided search results (URL and content). You must
only use information from the provided search results. Use an unbiased and
journalistic tone. Combine search results together into a coherent answer. Do not
repeat text. Cite search results using [${{number}}] notation. Only cite the most
relevant results that answer the question accurately. Place these citations at the end
of the sentence or paragraph that reference them - do not put them all at the end. If
different results refer to different entities within the same name, write separate
answers for each entity.

You should use bullet points in your answer for readability. Put citations where they apply
rather than putting them all at the end.

If there is nothing in the context relevant to the question at hand, just say "Hmm,
I'm not sure." Don't try to make up an answer.

For the Code: Produce complete, functional Python code that utilizes the Langchain framework to address the question or task
presented. Ensure that the code is:

Readable: Write clear, understandable code. Use descriptive variable names and adhere to Python's PEP 8 style
guidelines for maximum readability.

Efficient: Optimize for performance. Use efficient algorithms and data structures to ensure the code runs
effectively and conservatively uses resources.

Error-Handled: Incorporate error handling to manage and anticipate potential issues that might arise during
execution. Use try-except blocks where appropriate and validate input data.

Tested: Include necessary assertions or print statements to demonstrate the functionality and correctness
of the code. Ensure that the code produces expected results when executed.

Complete: Ensure the code can be copied directly into a Python project with no modifications needed.
It should be self-contained, specifying all necessary imports and dependencies.

Langchain-Specific: Utilize the appropriate classes, functions, and methods from the Langchain
framework. Clearly demonstrate how Langchain's features and capabilities are applied to solve the
given problem or task.

Context-Aware: Use the provided context effectively. Reference and utilize data, variables, or
insights derived from the 'context' section to inform the code's logic and functionality. \

Comment-Free Execution: Avoid using comment blocks as placeholders for code. Ensure all functional
parts of the code are executable statements and not comments.

If the question or task does not provide enough information for a complete script, or if it is outside the scope of
the Langchain framework, clearly state that a complete working script cannot be provided as requested, explain
the reasons and provide suggestions and/or alternatives.

Anything between the following context html blocks is retrieved from a knowledge
bank, and is provided in addition to the conversation with the user. <context> {context} <context/>

Anything between the following Code_Block html blocks were provided by user as integral and shall be
treated as essential to your response. Each block of code is enclosed within a pair of markdowns triple backticks (```) <Code_Block> {Code_Block} <Code_Block/>

REMEMBER: -If there is no relevant information within the context, just say "Hmm, I'm
not sure." Don't try to make up an answer. Anything between the preceding 'context'
html blocks is retrieved from a knowledge bank, not part of the conversation with the
user. -The goal is to provide code that is ready to be integrated and used in a project,
demonstrating the practical application and power of the Langchain framework in solving real-world problems.
Anything between the preceding 'Code_Block' html blocks were provided by user as integral and shall be
treated as essential to your response. """
)



from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable

from pydantic import BaseModel, Field
from termcolor import colored

from . import prompting
from .llms import ask_oai
from .message import MESSAGES_TYPE, last_user_message, user_message
from .prompting.prompt_fns import (
    create_template_functions,
    process_prompt,
)
from .tools import use_tool
from .utils import (
    flatten_list,
    get_function_name,
    get_param_names,
    get_required_param_names,
)

TEMPLATE_FUNCTIONS = create_template_functions([prompting])
ASKER = ask_oai


def check_exit(messages: MESSAGES_TYPE = None) -> bool:
    if not messages:
        return False
    last_content = messages[-1]["content"]
    if isinstance(last_content, str) and last_content.lower() == "exit":
        messages.pop(-1)
        return True
    last_user_msg = last_user_message(messages)
    if last_user_msg:
        last_user_content = last_user_msg["message"]["content"]
        if isinstance(last_user_content, str) and last_user_content.lower() == "exit":
            messages.pop(last_user_msg["index"])
            return True
    return False


def run_task(
    task: Any, assets: dict | None = None, asset_key: str | None = None
) -> dict:
    if asset_key and asset_key == "skip":
        return assets
    assets = assets or OrderedDict()
    assets.setdefault("messages", [])
    if isinstance(task, dict):
        return run_dict_task(task, assets)
    return run_list_task(task, assets, asset_key)


def run_dict_task(task: dict, assets: dict) -> dict:
    for asset_key, subtask in task.items():
        assets = run_task(
            task=subtask,
            assets=assets,
            asset_key=asset_key,
        )
    return assets


def run_list_task(task, assets: dict, asset_key: str | None = None) -> dict:
    task = [task] if not isinstance(task, list) else task
    task = flatten_list(task)
    for subtask in task:
        if check_exit(assets["messages"]):
            break
        if isinstance(subtask, dict):
            current_asset_key, subtask = list(subtask.items())[0]
        else:
            current_asset_key = asset_key
        assets = process_subtask(subtask, assets, current_asset_key)
    return assets


def process_subtask(subtask, assets: dict, asset_key: str | None = None) -> dict:
    if callable(subtask):
        assets = use_tool(
            tool=subtask,
            assets=assets,
            asset_key=asset_key,
        )
    elif isinstance(subtask, (str, Path)):
        task_content = process_prompt(subtask, template_functions=TEMPLATE_FUNCTIONS)
        assets["messages"].append(user_message(content=task_content))
        if asset_key:
            assets[asset_key] = task_content
    else:
        raise TypeError(f"Invalid task type: {type(subtask)}")
    return assets


def ask(*tasks, assets: dict | None = None) -> dict:
    assets = assets or OrderedDict()
    assets.setdefault("messages", [])
    messages = assets["messages"]
    tasks = list(tasks)
    # print(f"\n\nASK TASKS: {tasks}\n\n")
    # print(f"\n\nASK MESSAGES: {messages}\n\n")
    if len(messages) > 0 and messages[-1]["role"] == "user":
        if tasks[0] != messages[-1]["content"]:
            tasks.insert(0, messages.pop(-1)["content"])
    for task in tasks:
        try:
            # if isinstance(task, Task):
            #     if task.condition(assets):
            #         assets = run_task(task=task.subtasks, assets=assets)
            # else:
            assets = run_task(task=task, assets=assets)
        except Exception as e:
            print(colored(f"\n\nASK ERROR: {e}\n\n", "red"))
        if check_exit(assets["messages"]):
            break
    return assets


def default_assets(assets: dict = None) -> dict:
    assets = assets or OrderedDict()
    assets.setdefault("messages", [])
    return assets


class Task(BaseModel):
    name: str = ""
    condition: Callable = Field(default_factory=lambda: lambda x: True)
    task: Any

    def do_task(self, assets: dict | None = None, **kwargs) -> dict:
        assets = default_assets(assets=assets)
        result_dict = {}
        if isinstance(self.task, (str, Path)):
            task_result = process_prompt(
                prompt=self.task, template_functions=TEMPLATE_FUNCTIONS
            )
            assets["messages"].append(user_message(content=task_result))
            result_name = self.name or "_".join(str(self.task).split()[:2]) + "_result"
            result_dict[result_name] = task_result
        elif callable(self.task):
            params = {"assets": assets, **assets, **kwargs}
            task_name = get_function_name(self.task)
            task_params = get_param_names(self.task)
            task_required_params = get_required_param_names(self.task)
            result_name = self.name or f"{task_name}_result"
            if "kwargs" in task_params:
                call_params = params
            else:
                call_params = {k: v for k, v in params.items() if k in task_params}
            if call_params or not task_required_params:
                curr_messages = deepcopy(assets["messages"])
                try:
                    task_result = self.task(**call_params)
                    result_dict[result_name] = task_result
                    if (
                        assets["messages"] != curr_messages
                    ):  # the tool modified the messages
                        if (
                            task_result == assets["messages"]
                        ):  # the tool returned the modified messages
                            result_dict = {result_name: task_result[-1]["content"]}
                        elif (
                            task_result != assets["messages"][-1]["content"]
                        ):  # the tool modified the messages and also returned something else
                            result_dict = {
                                result_name: task_result,
                                f"{result_name}_message": assets["messages"][-1][
                                    "content"
                                ],
                            }
                    elif task_result == assets["messages"]:  # nothing was modified
                        result_dict = {}
                except Exception as e:
                    error_msg = f"ERROR WITH TASK: {task_name}\nERROR: {e}"
                    print(colored(f"\n\n{error_msg}\n\n", "red"))
        else:
            raise TypeError(f"Invalid task type: {type(self.task)}")
        assets.update(result_dict)
        return assets

    def __call__(self, assets: dict | None = None, **kwargs) -> dict:
        assets = default_assets(assets=assets)
        try:
            condition = self.condition(assets=assets, task=self.task)
        except Exception:
            condition = self.condition(assets=assets)
        if condition:
            assets = self.do_task(assets=assets, **kwargs)
        return assets


class Tasks(BaseModel):
    tasks: list[Task] = Field(default_factory=list)
    condition: Callable = Field(default_factory=lambda: lambda x: True)

    def do_tasks(self, assets: dict | None = None, **kwargs) -> dict:
        assets = default_assets(assets=assets)
        if self.condition(assets):
            for task in self.tasks:
                if not isinstance(task, Task):
                    task = Task(task=task)
                assets = task(assets=assets, **kwargs)
        return assets


class Asker(BaseModel):
    assets: dict = Field(default_factory=default_assets)
    tasks: Tasks = Field(default_factory=Tasks)

    def __call__(self, assets: dict | None = None, **kwargs) -> dict:
        assets = default_assets(assets=assets)
        assets = self.tasks.do_tasks(assets=assets, **kwargs)
        return assets



import re
from pathlib import Path

import pandas as pd
from chromadb import Collection as ChromaCollection
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import UnstructuredPDFLoader
from langchain_core.documents import Document as LCDocument

from .chroma import (
    CHROMA_DEVICE,
    CHROMA_DIR,
    CHROMA_EMBEDDING_MODEL,
    chroma_collection,
    lc_docs_to_chroma_docs,
)
from .utils import resolve_data_path

CHUNK_SIZE = 100
CHUNK_OVERLAP = 10
SEPARATORS = ["."]
CHROM_COLLECTION_NAME = "pdf_collection"
CHROMA_DELETE_EXISTING = False
PARENTS_DF_NAME = "parents_df"
CHILDREN_COLLECTION_NAME = "children_collection"


def pdf_pages(
    pdf_file: str,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list[str] = SEPARATORS,
) -> list[LCDocument]:
    page_loader = UnstructuredPDFLoader(pdf_file, mode="paged")
    if chunk_size:
        page_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            separators=separators,
            keep_separator=False,
        )
        pages = page_loader.load_and_split(text_splitter=page_splitter)
    else:
        pages = page_loader.load()
    return pages


def pdf_to_collection(
    pdf_file: str,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list[str] = SEPARATORS,
    collection_name: str = CHROM_COLLECTION_NAME,
    persistent_dir: str = CHROMA_DIR,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
    embedding_model: str = CHROMA_EMBEDDING_MODEL,
    device: str = CHROMA_DEVICE,
) -> ChromaCollection:
    pages = pdf_pages(
        pdf_file=pdf_file,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=separators,
    )
    collection = chroma_collection(
        name=collection_name,
        persistent_dir=persistent_dir,
        delete_existing=delete_existing,
        embedding_model=embedding_model,
        device=device,
    )
    ids, docs, metadatas = lc_docs_to_chroma_docs(pages)
    collection.add(ids=ids, documents=docs, metadatas=metadatas)
    return collection


def pdf_pages_to_df(
    pdf_file: str,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list[str] = SEPARATORS,
) -> pd.DataFrame:
    pages = pdf_pages(
        pdf_file,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=separators,
    )
    df = pd.DataFrame(
        [
            {
                "page_number": page.metadata.get("page_number", 0),
                "page_content": page.page_content,
                "filename": page.metadata["filename"],
            }
            for page in pages
        ]
    )
    return df


def pdf_to_family_collection(
    data_path: str | list[str] | None = None,
    parent_pages: list[LCDocument] | None = None,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list[str] = SEPARATORS,
    parents_df_name: str = PARENTS_DF_NAME,
    children_collection_name: str = CHILDREN_COLLECTION_NAME,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
) -> tuple[ChromaCollection, pd.DataFrame]:
    assert (
        data_path or parent_pages
    ), "Either data_path or parent_pages must be provided."
    children_collection = chroma_collection(
        name=children_collection_name, delete_existing=delete_existing
    )
    page_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=separators,
        keep_separator=False,
    )
    parent_dfs = []
    if data_path:
        data_list = resolve_data_path(data_path)
    else:
        data_list = [parent_pages]
    for data in data_list:
        if isinstance(data, str):
            if Path(data).suffix != ".pdf":
                continue
            parent_pages = pdf_pages(data)
        else:
            parent_pages = data
        parent_df = pd.DataFrame(
            [
                {
                    "page_number": page.metadata.get("page_number", 0),
                    "page_content": re.sub(r"\n+", "\n", page.page_content),
                    "filename": page.metadata["filename"],
                }
                for page in parent_pages
            ]
        )
        parent_dfs.append(parent_df)

        children = page_splitter.split_documents(parent_pages)
        children_ids, children_texts, children_metadatas = lc_docs_to_chroma_docs(
            children
        )
        print(f"\n\nCHILDREN IDS: {children_ids}\n\n")
        children_collection.add(
            ids=children_ids, documents=children_texts, metadatas=children_metadatas
        )
    parents_df = pd.concat(parent_dfs).reset_index(drop=True)
    parents_df_name: Path = Path(parents_df_name).with_suffix(".csv")
    if not delete_existing and parents_df_name.exists():
        existing_parents_df = pd.read_csv(parents_df_name)
        parents_df = pd.concat([existing_parents_df, parents_df]).reset_index(drop=True)
    parents_df.to_csv(parents_df_name, index=False)
    return children_collection, parents_df



from typing import Callable
from uuid import uuid4

import chromadb
from chromadb import Collection as ChromaCollection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from langchain_core.documents import Document as LCDocument
from termcolor import colored

CHROMA_EMBEDDING_MODEL = "multi-qa-mpnet-base-cos-v1"
CHROMA_DIR = "chroma_dir"
CHROMA_DEVICE = "cuda"
CHROMA_DELETE_EXISTING = False


def chroma_collection(
    name: str,
    persistent_dir: str = CHROMA_DIR,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
    embedding_model: str = CHROMA_EMBEDDING_MODEL,
    device: str = CHROMA_DEVICE,
) -> ChromaCollection:
    chroma_client = chromadb.PersistentClient(path=persistent_dir)
    if delete_existing:
        try:
            chroma_client.delete_collection(name)
        except Exception as e:
            print(colored(f"Error deleting collection named {name}: {e}", "red"))
            pass
    embedding_function = SentenceTransformerEmbeddingFunction(
        model_name=embedding_model, device=device
    )
    collection = chroma_client.get_or_create_collection(
        name, embedding_function=embedding_function
    )
    return collection


def id_from_lc_doc(doc: LCDocument, doc_idx=None, **kwargs) -> str | None:
    """Get a unique id from an LCDocument."""

    try:
        file_name = doc.metadata["filename"]
        doc_id = f"{file_name}_{doc.metadata.get('page_number', 0)}"
        if doc_idx is not None:
            doc_id += f"_{doc_idx}"
        return doc_id + f"_{uuid4()}"
    except Exception as e:
        print(colored(f"Error getting id from LCDocument: {e}", "red"))
        return None


def lc_docs_to_chroma_docs(
    docs: list[LCDocument], id_fn: Callable = id_from_lc_doc
) -> tuple[list[str], list[str], list[dict]]:
    """Convert a list of LCDocuments to Chroma docs with ids and metadata."""

    chroma_ids = []
    chroma_docs = []
    chroma_metadatas = []
    for i, doc in enumerate(docs):
        chroma_docs.append(doc.page_content)
        chroma_metadatas.append(
            {
                k: v
                for k, v in doc.metadata.items()
                if type(v) in [str, int, float, bool]
            }
        )
        chroma_ids.append(id_fn(doc=doc, doc_idx=i) or f"id_{i}")
    return chroma_ids, chroma_docs, chroma_metadatas



from ..utils import deindent
from .edu_prompt_strs import mcq_prompt, slides_prompt
from .prompt_fns import titles_w_content_template


def course_template(course_name: str, prefix: str = "", **extra) -> str:
    prefix = (
        deindent(prefix)
        or "You are a friendly and helpful instructional coach helping teachers create engaging and effective lessons for their students."
    )
    if course_name:
        prefix += f"\nThe course you are currently coaching is: {course_name}."
    return titles_w_content_template(extra, prefix=prefix)


def lesson_template(lesson_name: str, **lesson_reqs) -> str:
    prompt = f"Today's lesson is: {lesson_name}.\n"
    return titles_w_content_template(lesson_reqs, prefix=prompt)


def slides_prompt_template(n_bullets: int = 4) -> str:
    return (
        slides_prompt
        + f"\nYou MUST give me at least {n_bullets} bullet points for per slide."
    )


def mcq_quiz_prompt_template(n_questions: int = 10) -> str:
    return mcq_prompt + f"\nYou MUST give me {n_questions} questions."



from copy import deepcopy
from enum import Enum

import instructor
import litellm
import tiktoken
from dsp.modules.lm import LM as DSPyLM
from litellm import completion
from openai import OpenAI
from termcolor import colored

from .message import assistant_message
from .utils import retry_attempts

litellm.drop_params = True


class ModelName(str, Enum):
    GPT_3 = "gpt-3.5-turbo"
    GPT_4 = "gpt-4-turbo-preview"
    GEMINI = "gemini/gemini-pro"
    CLAUDE2 = "claude-2.1"
    CLAUDE3 = "claude-3-haiku-20240307"
    MISTRAL = "anyscale/mistralai/Mistral-7B-Instruct-v0.1"
    MIXTRAL = "anyscale/mistralai/Mixtral-8x7B-Instruct-v0.1"


GPT_MODEL = ModelName.GPT_3
GEMINI_MODEL = ModelName.GEMINI


def dspy_prompt(lm: DSPyLM) -> str:
    n = 1
    skip = 0
    provider: str = lm.provider
    last_prompt = None
    printed = []
    n = n + skip
    for x in reversed(lm.history[-100:]):
        prompt = x["prompt"]
        if prompt != last_prompt:
            if provider == "clarifai" or provider == "google":
                printed.append(
                    (
                        prompt,
                        x["response"],
                    ),
                )
            else:
                printed.append(
                    (
                        prompt,
                        x["response"].generations
                        if provider == "cohere"
                        else x["response"]["choices"],
                    ),
                )
        last_prompt = prompt
        if len(printed) >= n:
            break
    history_str = ""
    for idx, (prompt, choices) in enumerate(reversed(printed)):
        if (n - idx - 1) < skip:
            continue
        history_str += prompt
        text = ""
        if provider == "cohere":
            text = choices[0].text
        elif provider == "openai" or provider == "ollama":
            text = " " + lm._get_choice_text(choices[0]).strip()
        elif provider == "clarifai":
            text = choices
        elif provider == "google":
            text = choices[0].parts[0].text
        else:
            text = choices[0]["text"]
        history_str += text
        if len(choices) > 1:
            history_str += f" \t (and {len(choices)-1} other completions)"
    return history_str


def oai_response(response) -> str:
    try:
        return response.choices[0].message.content
    except Exception:
        return response


def ask_litellm(
    messages: list[dict[str, str]],
    model: str = GEMINI_MODEL,
    json_mode: bool | None = None,
) -> tuple:
    try:
        if json_mode is None and "json" in messages[-1]["content"].lower():
            response_format = {"type": "json_object"}
        else:
            response_format = None
        # response_format = {
        #     "type": (
        #         "json_object"
        #         if json_mode is None and "json" in messages[-1]["content"].lower()
        #         else "text"
        #     )
        # }
        answer = oai_response(
            completion(
                messages=deepcopy(messages),
                model=model,
                response_format=response_format,
            )
        )
        messages.append(assistant_message(content=answer))
        return answer, messages
    except Exception as e:
        print(colored(f"\n\n{model} ERROR: {e}\n\n", "red"))
        return None, messages


def count_gpt_tokens(text: str, model: str = GPT_MODEL) -> int:
    encoding = tiktoken.encoding_for_model(model)
    return len(encoding.encode(text))


def ask_oai(
    messages: list[dict[str, str]],
    model: str = GPT_MODEL,
    response_model=None,
    temperature: float = 0.3,
    max_tokens: int | None = None,
    patch_mode=instructor.Mode.TOOLS,
    attempts: int = 1,
    api_key: str | None = None,
    base_url: str | None = None,
) -> tuple:
    """Ask OpenAI to generate a response to the given messages."""
    # print(f"\n\nMESSAGES: {messages}\n\n")
    gpt = instructor.patch(OpenAI(api_key=api_key, base_url=base_url), mode=patch_mode)
    try:
        answer = gpt.chat.completions.create(
            messages=deepcopy(messages),
            model=model,
            response_model=response_model,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=retry_attempts(attempts=attempts),
        )
    except Exception as e:
        print(colored(f"\n\n{model} ERROR: {e}\n\n", "red"))
        return None, messages
    if not response_model:
        answer = oai_response(answer)
        messages.append(assistant_message(content=answer))
        return answer, messages
    messages.append(assistant_message(content=answer.model_dump_json()))
    return answer, messages



import json
import re
from collections import Counter, OrderedDict
from copy import deepcopy
from typing import Any, Callable

import pandas as pd
import torch
from langchain.tools.brave_search.tool import BraveSearch
from langchain.tools.ddg_search.tool import DuckDuckGoSearchRun
from langchain.utilities.python import PythonREPL
from sentence_transformers import CrossEncoder
from termcolor import colored

from .chroma import ChromaCollection
from .g_apis.quiz import add_question, create_quiz
from .g_apis.slides import add_slide, create_presentation
from .message import MESSAGES_TYPE, assistant_message, user_message
from .prompting.prompt_fns import (
    chat_template,
    generic_summary_template,
    rag_template,
    titles_w_content_template,
)
from .utils import (
    count_tokens,
    get_function_name,
    get_param_names,
    get_required_param_names,
    token_count_to_line_count,
)

CROSS_ENCODER_MODEL = "cross-encoder/ms-marco-MiniLM-L-6-v2"
MAX_TOKENS = 25_000


def use_tool(
    tool: Callable,
    assets: dict | None = None,
    asset_key: str | None = None,
    **kwargs,
) -> dict:
    assets = OrderedDict(assets) or OrderedDict()
    assets.setdefault("messages", [])
    params = {**assets, **kwargs}
    tool_name = get_function_name(tool)
    tool_params = get_param_names(tool)
    tool_required_params = get_required_param_names(tool)
    # print(f"\n\nTOOL: {tool_name}\n\n")
    if "kwargs" in tool_params:
        call_params = params
    else:
        call_params = {k: v for k, v in params.items() if k in tool_params}
    if "assets" in tool_params:
        call_params["assets"] = assets
    if call_params or not tool_required_params:
        curr_messages = deepcopy(assets["messages"])
        try:
            ret = tool(**call_params)
        except Exception as e:
            error_msg = f"ERROR WITH TOOL: {tool_name}\nERROR: {e}"
            print(colored(f"\n\n{error_msg}\n\n", "red"))
            # assets["messages"].append(
            #     user_message(
            #         content=f"ERROR WITH TOOL: {get_function_source(tool)}\nERROR: {e}"
            #     )
            # )
            ret = None
        try:
            if ret is not None:
                asset_key = asset_key or f"{tool_name}_result"
                ret_dict = {asset_key: ret}
                # print(f"\n\nRET: {ret}\n\n")
                # print(f"\n\nASSETS: {assets}\n\n")
                # print(f"\n\nCURR MESSAGES: {curr_messages}\n\n")
                if (
                    assets["messages"] != curr_messages
                ):  # the tool modified the messages
                    if (
                        ret == assets["messages"]
                    ):  # the tool returned the modified messages
                        ret_dict = {asset_key: ret[-1]["content"]}
                    elif (
                        ret != assets["messages"][-1]["content"]
                    ):  # the tool modified the messages and also returned something else
                        ret_dict = {
                            asset_key: ret,
                            f"{asset_key}_message": assets["messages"][-1]["content"],
                        }
                elif ret == assets["messages"]:  # nothing was modified
                    ret_dict = {}
                assets.update(ret_dict)
        except Exception as e:
            error_msg = f"ERROR WITH TOOL RESULT: {tool_name}\nERROR: {e}"
            print(colored(f"\n\n{error_msg}\n\n", "red"))
            # assets["messages"].append(
            #     user_message(
            #         content=f"ERROR WITH TOOL RESULT: {get_function_source(tool)}\nERROR: {e}"
            #     )
            # )
    return assets


def format_brave_snippets(snippets, start=1):
    ret = ""
    for i, s in enumerate(snippets, start=start):
        title = s["title"]
        snippet = s["snippet"]
        snippet = re.sub(
            r"<strong>(.*?)</strong>", lambda x: x.group(1).upper(), snippet
        )
        ret += f"[{i}] {title}\n"
        ret += f"{snippet}\n\n"
    return ret


# @process_messages
def brave_search(
    messages: MESSAGES_TYPE = None, query: str = "", count: int = 5
) -> MESSAGES_TYPE:
    """Query the Brave web search API and add the results to the messages list."""
    if not messages:
        return []
    if not query:
        return messages
    search = BraveSearch.from_api_key(
        api_key="BSAYt3qZcqqKxxuFTEZeA7BL0EiRmqs", search_kwargs={"count": count}
    )
    query = query or messages[-1]["content"]
    # print(f"\n\nQUERY: {query}\n\n")
    res = json.loads(search.run(query))
    # print(f"\n\nBRAVE SEARCH RESULTS: {res}\n\n")
    messages.append(
        user_message(
            content=titles_w_content_template(
                {"Up to date Web Search Results For You": format_brave_snippets(res)}
            )
        )
    )
    return messages


# @process_messages
def ddg_search(messages: MESSAGES_TYPE = None, query: str = "") -> MESSAGES_TYPE:
    """
    Query the DuckDuckGo web search API and add the results to the messages list.
    If there are no good results, try Brave Search.
    """

    if not messages:
        return []
    if not query:
        return messages
    search = DuckDuckGoSearchRun()
    query = query or messages[-1]["content"]
    print(f"\n\nQUERY: {query}\n\n")
    res = search.run(query)
    if "No good DuckDuckGo Search Result was found" in res:
        return brave_search(messages=messages, query=query)
    messages.append(
        user_message(
            content=titles_w_content_template(
                {"Up to date Web Search Results For Your Task": res}
            )
        )
    )
    return messages


# @process_messages
def upload_slides(
    messages: MESSAGES_TYPE,
    service: Any,
    name: str = "DREAMAI SLIDES",
    title: str = "DREAMAI",
    subtitle: str = "",
) -> str | None:
    if not messages:
        return None
    course = json.loads(messages[-1]["content"])
    presentation_id = create_presentation(
        service=service,
        name=name,
        title=title,
        subtitle=subtitle,
    )
    [
        add_slide(
            service=service,
            presentation_id=presentation_id,
            title_text=title,
            body_text=bullets,
        )
        for title, bullets in zip(course["titles"], course["bullets"])
    ]
    return presentation_id


# @process_messages
def upload_quiz(
    messages: MESSAGES_TYPE,
    service: Any,
    name: str = "DREAMAI QUIZ",
    title: str = "DREAMAI",
) -> str | None:
    if not messages:
        return None
    quiz = json.loads(messages[-1]["content"])
    quiz_id = create_quiz(
        service=service,
        name=name,
        title=title,
    )
    [
        add_question(
            question=question,
            answers=answers,
            correct_index=correct_index,
            service=service,
            quiz_id=quiz_id,
        )
        for question, answers, correct_index in zip(
            quiz["questions"], quiz["answers"], quiz["correct_indices"]
        )
    ]
    return quiz_id


# @process_messages
def user_feedback(
    messages: MESSAGES_TYPE = None,
    asssistant_query: str = "",
    history_len: int = 2,
) -> MESSAGES_TYPE:
    """Get user feedback on a list of messages."""
    if not messages:
        return []
    if history_len > 0:
        print(chat_template(messages[-history_len:], with_colors=True))
    if asssistant_query:
        messages.append(assistant_message(content=asssistant_query))
    user_response = input(f"{asssistant_query}\n")
    messages.append(user_message(content=user_response))
    return messages


# @process_messages
def run_code(
    messages: MESSAGES_TYPE = None,
    code: str = "",
    _globals: dict | None = None,
    _locals: dict | None = None,
) -> MESSAGES_TYPE:
    """Run code in a Python REPL and add the result to the messages list."""
    if not messages:
        return []
    python_repl = PythonREPL(_globals=_globals, _locals=_locals)
    if not code:
        code = next(
            (
                message["content"]
                for message in reversed(messages)
                if "```python" in message["content"]
            ),
            None,
        )
        if not code:
            return messages
    try:
        code = code.split("```python")[1].split("```")[0].strip()
        code_result = python_repl.run(code)
    except Exception as e:
        code_result = str(e)
    if code_result:
        messages.append(assistant_message(content=code_result))
    return messages


# @process_messages
def chroma_retriever(
    messages: MESSAGES_TYPE,
    collection: ChromaCollection,
    n_results: int = 10,
    **kwargs,
) -> MESSAGES_TYPE:
    """Query the Chroma collection with the last message as the query text."""
    if not messages:
        return []
    query_text = messages[-1]["content"]
    query_res = collection.query(query_texts=query_text, n_results=n_results)
    messages.append(user_message(content=rag_template(info=query_res["documents"][0])))
    return messages


def rerank_chroma_results(
    query_text: str,
    results: dict,
    n_rerank: int = 40,
    cross_encoder_model: str = CROSS_ENCODER_MODEL,
) -> dict:
    """Rerank the results with a cross-encoder model."""
    if n_rerank:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        cross_encoder = CrossEncoder(cross_encoder_model, device=device)
        pairs = [[query_text, doc] for doc in results["documents"][0]]
        scores = cross_encoder.predict(pairs)
        scores_idx = sorted(range(len(scores)), key=lambda i: scores[i], reverse=True)
        results = {k: [[v[0][i] for i in scores_idx]] for k, v in results.items()}
    return results


def chroma_family_retreiver_2(
    query_text: str,
    children_collection: ChromaCollection,
    parents_df: pd.DataFrame | None = None,
    n_children: int = 100,
    n_rerank: int = 50,
    children_per_parent: int = 2,
    metadata_filter: dict | None = None,
    only_get: bool = False,
) -> str:
    if only_get and metadata_filter is not None:
        children = children_collection.get(where=metadata_filter)
        children_ids = children["ids"]
        children_docs = children["documents"]
        children_metadatas = children["metadatas"]
        print(f"\n\nQUERIED CHILDREN: {children['ids']}\n\n")
    elif not query_text:
        return ""
    else:
        print(f"\n\nQUERY TEXT: {query_text}\n\n")
        children = children_collection.query(
            query_texts=query_text, n_results=n_children, where=metadata_filter
        )
        print(f"\n\nQUERIED CHILDREN: {children['ids'][0]}\n\n")
        children = rerank_chroma_results(
            results=children, query_text=query_text, n_rerank=n_rerank
        )
        children_ids = children["ids"][0]
        children_docs = children["documents"][0]
        children_metadatas = children["metadatas"][0]
    print(f"\n\nCHILDREN: {children_ids}\n\n")
    if len(parents_df) == 0:
        parents = children_docs
    else:
        children_file_pages = [
            (metadata["filename"], metadata.get("page_number", 0))
            for metadata in children_metadatas
        ]
        children_file_page_counts = Counter(children_file_pages)
        parents_file_pages = sorted(
            [
                file_page
                for file_page, count in children_file_page_counts.items()
                if count >= children_per_parent
            ],
            key=lambda x: x[1],
        )
        print(f"\n\nPARENTS FILE PAGES: {parents_file_pages}\n\n")
        parents_file_pages = parents_file_pages or children_file_pages
        parents = parents_df[
            parents_df.apply(
                lambda row: (row["filename"], row["page_number"]) in parents_file_pages,
                axis=1,
            )
        ]["page_content"].tolist()
    return "\n\n".join(parents) if parents else ""


# @process_messages
def chroma_family_retreiver(
    messages: MESSAGES_TYPE,
    children_collection: ChromaCollection,
    parents_df: pd.DataFrame | None = None,
    n_children: int = 100,
    n_rerank: int = 50,
    children_per_parent: int = 2,
    metadata_filter: dict | None = None,
    only_get: bool = False,
) -> MESSAGES_TYPE:
    """Query the children collection with the last message as the query text and then retrieve the parents."""
    if only_get and metadata_filter is not None:
        children = children_collection.get(where=metadata_filter)
        children_ids = children["ids"]
        children_docs = children["documents"]
        children_metadatas = children["metadatas"]
        print(f"\n\nQUERIED CHILDREN: {children['ids']}\n\n")
        messages = messages or []
    elif not messages:
        return []
    else:
        query_text = messages[-1]["content"]
        print(f"\n\nQUERY TEXT: {query_text}\n\n")
        children = children_collection.query(
            query_texts=query_text, n_results=n_children, where=metadata_filter
        )
        print(f"\n\nQUERIED CHILDREN: {children['ids'][0]}\n\n")
        children = rerank_chroma_results(
            results=children, query_text=query_text, n_rerank=n_rerank
        )
        children_ids = children["ids"][0]
        children_docs = children["documents"][0]
        children_metadatas = children["metadatas"][0]
    print(f"\n\nCHILDREN: {children_ids}\n\n")
    if len(parents_df) == 0:
        parents = children_docs
    else:
        children_file_pages = [
            (metadata["filename"], metadata.get("page_number", 0))
            for metadata in children_metadatas
        ]
        children_file_page_counts = Counter(children_file_pages)
        parents_file_pages = sorted(
            [
                file_page
                for file_page, count in children_file_page_counts.items()
                if count >= children_per_parent
            ],
            key=lambda x: x[1],
        )
        print(f"\n\nPARENTS FILE PAGES: {parents_file_pages}\n\n")
        parents_file_pages = parents_file_pages or children_file_pages
        parents = parents_df[
            parents_df.apply(
                lambda row: (row["filename"], row["page_number"]) in parents_file_pages,
                axis=1,
            )
        ]["page_content"].tolist()
    if parents:
        messages.append(user_message(content=rag_template(info=parents)))
    return messages


# @process_messages
def summarize_messages(
    messages: MESSAGES_TYPE,
    asker: Callable,
    max_tokens: int = MAX_TOKENS,
    **kwargs,
) -> MESSAGES_TYPE:
    """Summarize the chat if it is longer than max_tokens."""
    if not messages:
        return []
    if count_tokens(chat_template(messages)) <= max_tokens:
        return messages
    line_count = token_count_to_line_count(max_tokens)
    system_msg = None
    user_msg = None
    if messages[0]["role"] == "system":
        system_msg = messages.pop(0)
    if len(messages) > 1 and messages[-1]["role"] == "user":
        other_messages = messages[:-1]
        user_msg = messages[-1]
        other_messages_token_count = count_tokens(chat_template(other_messages))
        user_msg_token_count = count_tokens(user_msg["content"])
        if user_msg_token_count < other_messages_token_count:
            user_msg = messages.pop(-1)
        else:
            user_msg = None

    if not messages:
        return [message for message in [system_msg, user_msg] if message]

    chat_str = chat_template(messages)
    summary_prompt = titles_w_content_template(
        titles_dict={
            "conversation": chat_str,
            "task": generic_summary_template(line_count=line_count),
        },
        # suffix=generic_summary_template(line_count=line_count),
    )
    messages = [user_message(content=summary_prompt)]
    print(f"\n\nSUMMARY PROMPT: {summary_prompt}\n\n")
    messages = asker(messages=messages)
    chat_summary = messages[-1]["content"]
    messages = [
        user_message(
            content=titles_w_content_template(
                titles_dict={"chat_summary": chat_summary}
            )
        )
    ]
    if system_msg:
        messages.insert(0, system_msg)
    if user_msg:
        messages.append(user_msg)
    return messages


# @process_messages
def summarize_then_ask(
    messages: MESSAGES_TYPE,
    asker: Callable,
    max_tokens: int = MAX_TOKENS,
    **kwargs,
) -> MESSAGES_TYPE:
    """Summarize the chat if it is longer than max_tokens and then ask a question."""
    if not messages:
        return []
    messages = summarize_messages(messages=messages, asker=asker, max_tokens=max_tokens)
    return asker(messages=messages)



from ..utils import deindent


notes_prompt = deindent(
    """
    Please create comprehensive lecture notes that cover all key topics and subtopics about the upcoming lesson. The notes should include:
    - Detailed explanations of each subject area.
    - Definitions and clarifications of relevant terminology.
    - Practical examples and case studies where applicable.
    - Steps and strategies highlighting the instructional focus.
    - Summaries of theoretical frameworks or models introduced.
    - Interactive activity guidelines and expected learning outcomes.
    - Instructions for homework assignments and expected deliverables.
    - Additional resources for expanded learning on the lesson's topics.
    The notes should be organized in a clear, logical manner, suitable for academic use, and supportive of student comprehension and engagement. Make sure the content is appropriately sectioned for ease of reading and study, ready for print and distribution.
    Make sure the notesare enough for the whole duration of the lesson.
    """
)

slides_prompt = deindent(
    """
    I need detailed lecture slides with bullet points for the upcoming lesson. Include the approximate duration for each slide. Make sure they add up to the total lesson duration if mentioned above.
    The slides must strike a balance between engagement and informativeness, designed to facilitate independent learning for students.
    Aim for clarity without sacrificing accuracy or detail. Craft explanations, analogies, and examples with simplicity, ensuring students can grasp the content without the presence of a teacher.
    Please generate the actual content for each slide rather than just providing titles and guidelines. Something I can copy-paste directly.
    We can iterate on the slides together to refine and enhance them, so perfection is not necessary in the initial draft.
    Your goal is to make the material accessible and understandable to students studying independently.
    """
)

slides_format_prompt = deindent(
    """
    Take this material and convert it into a JSON with the following format:
    {"titles": ["title1", "title2", "title3"], "bullets": [["bullet1", "bullet2"], ["bullet1", "bullet2"], ["bullet1", "bullet2"]]})
    """
)

slides_notes_prompt = deindent(
    "Generate the instructor notes for the slides. These notes will have the complete extra detailed content for each slide."
)

mcq_scoring_prompt = "We will make a quiz for this lesson. Design a scoring guide that outlines the specific criteria for evaluating student responses. Assume all quiz questions are MCQs."

mcq_prompt = deindent(
    """
    You are designing a series of multiple-choice questions as a quiz for the lesson. Each question should test the students' knowledge, understanding, and ability to apply key concepts from the course material. Follow these guidelines:
    1. **Topic Focus**: Each question should be based on a major topic or sub-topic covered in the lesson.
    2. **Question Clarity**: Formulate each question clearly and concisely, ensuring that the central focus of the question is straightforward and unambiguous.
    3. **Distractors**: Provide 3-4 answer choices for every question. The incorrect choices (distractors) should be plausible and reflect common misconceptions or errors related to the topic to challenge students to think critically.
    4. **Complexity**: Design questions that require more than just recall of information. Include questions that test students' ability to analyze, synthesize, and evaluate information.
    5. **Variety**: Mix different types of questions, such as fact-based, scenario-based, and application-based questions, to cover various levels of Bloom’s Taxonomy.
    """
)

quiz_prompt = deindent(
    """
    Building on the lecture slides, I now need a short quiz to assess students' understanding of the material.
    Design a quiz that covers key concepts presented in the slides, ensuring a balance between challenging questions and those that reinforce fundamental knowledge.
    The quiz should reflect the simplicity and clarity of the lecture content. Each question should be accompanied by clear explanations of the correct answers to aid students in their learning process.
    Feel free to propose an initial draft of the quiz, and we can collaborate on refining it to ensure it aligns with the learning objectives of the lesson.
    The goal is to create an effective assessment tool that reinforces understanding and promotes active learning.
    """
)

quiz_format_prompt = deindent(
    """
    Reformat the quiz material into a JSON with the following format:
    {"questions": ["question1", "question2", "question3"], "answers": [["answer1", "answer2"], ["answer1", "answer2"], ["answer1", "answer2"]], "correct_indices": [0, 1, 0]}
    """
)



from functools import partial
from typing import Any, Callable

from .llms import ModelName, ask_litellm
from .message import del_messages, set_user_role
from .prompting.edu_prompt_strs import (
    mcq_prompt,
    quiz_format_prompt,
    slides_format_prompt,
    slides_prompt,
)
from .prompting.prompt_strs import gen_rag_query_prompt, improve_prompt
from .tools import upload_quiz, upload_slides

ASKER = partial(ask_litellm, model=ModelName.GEMINI)


def re_prompt_recipe(prompt=improve_prompt, asker: Callable = ASKER):
    return [prompt, asker, partial(del_messages, idx=[-2, -3]), set_user_role]


improve_prompt_recipe = partial(re_prompt_recipe, prompt=improve_prompt)


def gen_rag_query_then_retrieve_recipe(retriever: Callable, asker: Callable = ASKER):
    return [gen_rag_query_prompt, asker, retriever, partial(del_messages, idx=[-2, -3])]


def slides_recipe(
    course_name: str,
    lesson_name: str,
    lesson_prompt: str,
    slides_service: Any,
    asker: Callable = ASKER,
    retriever: Callable | None = None,
):
    lesson_slides_creator = partial(
        upload_slides,
        name=f"{course_name}-{lesson_name}",
        title=lesson_name,
        subtitle="",
        service=slides_service,
    )
    recipe = [
        lesson_prompt,
        slides_prompt,
        "You MUST give me at least 4 bullet points for per slide.",
        asker,
        slides_format_prompt,
        {f"{'_'.join(lesson_name.split())}_slides": asker},
        {"slides_id": lesson_slides_creator},
    ]
    if retriever:
        recipe.insert(
            2, gen_rag_query_then_retrieve_recipe(retriever=retriever, asker=asker)
        )
    return recipe


def quiz_recipe(
    course_name: str, lesson_name: str, forms_service: Any, asker: Callable = ASKER
):
    lesson_quiz_creator = partial(
        upload_quiz,
        name=f"{course_name}-{lesson_name}",
        title=lesson_name,
        service=forms_service,
    )

    return [
        # partial(print, "Creating a quiz...⏳"),
        mcq_prompt,
        "You MUST give me at least 10 MCQs based on the lesson above.",
        # asker,
        quiz_format_prompt,
        {f"{'_'.join(lesson_name.split())}_quiz": asker},
        {"quiz_id": lesson_quiz_creator},
    ]


# rephrase_rag_query_tasks = partial(
# re_prompt_tasks,
# prompt=rephrase_rag_query_prompt,
# )



import inspect
import json
import os
from pathlib import Path
from typing import Callable

from chromadb import Collection as ChromaCollection
from termcolor import colored

from ..utils import clean_text, deindent
from .prompt_strs import path_selection_prompt


def create_template_functions(modules: list) -> dict[str, Callable]:
    template_functions = {}
    for module in modules:
        for name, fn in inspect.getmembers(module, inspect.isfunction):
            if name.endswith("_template"):
                template_functions[name] = fn
    return template_functions


def json_to_prompt(
    prompt_file: str | Path, template_functions: dict[str, Callable] | None = None
) -> str:
    if not os.path.exists(prompt_file := str(prompt_file)):
        return ""
    with open(prompt_file, "r") as f:
        prompt_data = json.load(f)
    # print(f"\n\nPROMPT DATA: {prompt_data}\n\n")
    if template_functions is None or "template_function" not in prompt_data:
        if "template_args" in prompt_data:
            template_args = prompt_data["template_args"]
        else:
            template_args = prompt_data
        if isinstance(template_args, dict) and "titles_dict" not in template_args:
            template_args = {"titles_dict": template_args}
        return titles_w_content_template(**template_args)
    template_function = template_functions[prompt_data["template_function"]]
    # print(f"\n\nTEMPLATE ARGS: {prompt_data['template_args']}\n\n")
    return template_function(**prompt_data["template_args"])


def txt_to_prompt(prompt_file: str | Path) -> str:
    if not os.path.exists(prompt_file := str(prompt_file)):
        return ""
    with open(prompt_file, "r") as f:
        prompt = f.read()
    return deindent(prompt)


def process_prompt(
    prompt: str | Path | list[str],
    template_functions: dict[str, Callable] | None = None,
) -> str:
    if not prompt:
        return ""
    if isinstance(prompt, list):
        prompt = "\n\n---\n\n".join(list(prompt))
    elif isinstance(prompt, (Path, str)):
        prompt = str(prompt)
        if prompt.endswith(".txt"):
            prompt = txt_to_prompt(prompt)
        elif prompt.endswith(".json"):
            prompt = json_to_prompt(prompt, template_functions=template_functions)
    return deindent(str(prompt))


def titles_w_content_template(
    titles_dict: dict[str, str | Path | list[str]] = {},
    prefix: str = "",
    suffix: str = "",
) -> str:
    prompt = deindent(prefix)
    for title, content in titles_dict.items():
        if content:
            title = " ".join(title.split("_")).strip().title()
            content = process_prompt(content)
            if content:
                if len(prompt) > 0:
                    prompt += "\n\n"
                prompt += deindent(f"## {title} ##\n\n{deindent(content)}")
    prompt += "\n\n" + deindent(suffix)
    return deindent(prompt)


def rag_template(info: str | list[str], prefix: str = "") -> str:
    return titles_w_content_template(
        titles_dict={"retrieved information": info},
        prefix=prefix or "Get help from this retrieved information for your response.",
    )


def message_template(message: dict[str, str]) -> str:
    assert "role" in message and ("content" in message or "parts" in message)
    role = message["role"]
    content = message["content"] if "content" in message else message["parts"]
    return deindent(f"## {role.upper()} ##\n\n{content}")


def chat_template(messages: list[dict[str, str]], with_colors=False) -> str:
    role_to_color = {
        "system": "cyan",
        "user": "cyan",
        "assistant": "green",
    }
    chat = []
    for message in messages:
        message_str = message_template(message)
        if with_colors:
            message_str = colored(message_str, role_to_color[message["role"]])
        chat.append(message_str)
    return deindent("\n\n".join(chat))


def path_selection_template(
    path_descs: list[str], paths_prompt: str = "", chat_history: str = ""
) -> str:
    return titles_w_content_template(
        {
            "Conversation History": deindent(chat_history),
            "Path Descriptions": titles_w_content_template(
                {f"Path {i}": path_desc for i, path_desc in enumerate(path_descs)}
            ),
        },
        prefix=deindent(paths_prompt or path_selection_prompt) + "\n\n",
    )


def gen_web_search_query_template(n_queries: int = 3):
    return deindent(
        f"""
        You are an expert at web search.
        I need {n_queries} web search queries based on the messages above. The messages are part of an ongoing conversation so they may have some unnecessary text like "Okay so tell me about...", "I want to know about...", "I'm curious about...", etc.
        Give me effective queries for a web search engine like Google, Bing, DuckDuckGo, etc.
        I will use your queries as they are, so please don't add any comments or notes or prefacing or ending statements.
        Thank you!
        """
    )


def generic_summary_template(line_count: int = 10) -> str:
    return deindent(
        f"""
    As a professional summarizer, create a concise and comprehensive summary of the provided text, be it an article, post, conversation, or passage, while adhering to these guidelines:
    1. Craft a summary that is detailed, thorough, in-depth, and complex, while maintaining clarity and conciseness.
    2. Incorporate main ideas and essential information, eliminating extraneous language and focusing on critical aspects.
    3. Keep all the important information in the summary.
    4. Rely strictly on the provided text, without including external information.
    5. Format the summary in paragraph form for easy understanding.
    6. It should be no more than {line_count} lines long.
    """
    )


def repeat_template(num_steps: int = 1) -> str:
    if num_steps == 1:
        return "Repeat the previous step that you did for this new prompt."
    else:
        return (
            f"Repeat the previous {num_steps} steps that you did for this new prompt."
        )


def first_n_pages_from_chroma_template(
    chroma_collection: ChromaCollection, n_pages: int = 20
) -> str:
    page_docs = chroma_collection.get(where={"page_number": {"$lt": n_pages}})[
        "documents"
    ]
    return titles_w_content_template(
        {f"First {n_pages} pages of the book": [clean_text(x) for x in page_docs]},
    )
